# Array to hold gridbox edge and center pressures
type JZPGridType
    pEdge::DArray
    pCenter::DArray
end

# Time specification
type TimeVecType
    year::Int
    month::Int
    day::Int
    hour::Int
    minute::Int
    second::Int
end

# Declare the StateMet type
type JZMetType
        # I-6 fields
        LWI::DArray             # LWI - Land/water/ice flags
        pSurf::DArray           # PS - Surface pressure (hPa)
        pSurfStart::DArray      #      (At start of dynamic timestep)
        pSurfEnd::DArray        #      (At end of dynamic timestep)
        pSeaLevel::DArray       # SLP - Sea-level pressure (hPa)
        O3DUCol::DArray         # TO3 - Total ozone column (DU)
        O3DUColStart::DArray    #       (At start of dynamic timestep)
        O3DUColEnd::DArray      #       (At end of dynamic timestep)
        O3DUTrop::DArray        # TTO3 - Tropospheric ozone column (DU)
        # A-3 fields
        sfcAlbedo::DArray       # ALBEDO - Surface albedo at 10 m (-)
        cldFrc2D::DArray        # CLDTOT - Cloud fraction at ground (-)
        evapFlux::DArray        # EVAP - Evapotranspiration flux (?)
        greenIdx::DArray        # GRN - "Greenness" index (?)
        wetRoot::DArray         # GWETROOT - Root soil wetness (-)
        wetTop::DArray          # GWETTOP - Topsoil wetness (-)
        sensFlux::DArray        # HFLUX - Sensible heat flux (W/m2)
        LAI::DArray             # LAI - Leaf area index (m2/m2)
        PARDiffuse::DArray      # PARDF - Photosynthetically-active radiation, diffuse (W/m2)
        PARDirect::DArray       # PARDR - Photosynthetically-active radiation, direct (W/m2)
        PBLDepth::DArray        # PBLH - Planetary boundary layer depth (m)
        prcpAnvSurf::DArray     # PRECANV - Anvil-source precipitation at ground (kg/m2/s)
        prcpTotSurf::DArray     # PRECTOT - Total precipitation at ground (kg/m2/s)
        prcpConSurf::DArray     # PRECCON - Convective precipitation at ground (kg/m2/s)
        prcpSnoSurf::DArray     # PRECSNO - Snow precipitation at ground (kg/m2/s)
        netRadLW::DArray        # LWGNET - Net longwave flux at ground (W/m2)
        upRadLW::DArray         # LWTUP - Outgoing longwave flux at ground (W/m2)
        specHum2m::DArray       # QV2M - Specific humidity at 2 m (kg/kg)
        netRadSW::DArray        # SWGNET - Net shortwave flux at groudn (W/m2)
        snowGeo::DArray         # SNODP - Geometric snow depth (m)
        snowH2O::DArray         # SNOMAS - GEOS-5 H2O equivalent snow depth (m)
        pTrop::DArray           # TROPP - Tropopause pressure (hPa, 2D)
        T2m::DArray             # T2M - Air temperature at 2 m (K)
        TSkin::DArray           # TSKIN - Radiance temperature (K)
        U10m::DArray            # U10M - U-wind at 10 m (m/s)
        UStar::DArray           # USTAR - Friction velocity (m/s)
        V10m::DArray            # V10M - V-wind at 10 m (m/s)
        Z0::DArray              # Z0 - Roughness height (m)
        EFlux::DArray           # EFLUX - Latent heat flux (W/m2)
        fracLand::DArray        # FRLAND - Land fraction (-)
        fracOcean::DArray       # FROCEAN - Ocean fraction (-)
        fracLake::DArray        # FRLAKE - Lake fraction (-)
        fracLandIce::DArray     # FRLANDICE - Land ice fraction (-)
        # A-6 fields
        cldFrc3D::DArray        # CLOUD - Cloud fraction (-)
        detrainFlux::DArray     # DTRAIN - Detrainment mass flux (kg/m2/s)
        convectFlux::DArray     # CMFMC - Convective mass flux (kg/m2/s)
        iceTendMoist::DArray    # DQIDTMST - Humidity tendency in ice processes (kg/kg/s)
        liqTendMoist::DArray    # DQLDTMST - Humidity tendency in moist processes (kg/kg/s)
        vapTendMoist::DArray    # DQVDTMST - Humidity tendency in vapor processes (kg/kg/s)
        prcpConForm::DArray     # DQRCON - Precipitation due to convective processes (kg/m2/s)
        prcpLrgForm::DArray     # DQRLSC - Precipitation due to large-scale processes (kg/m2/s)
        vapTend::DArray         # MOISTQ - Overall humidity tendency (derived) (kg/kg/s)
        tauCld::DArray          # OPTDEPTH - Optical depth, overall (-)
        tauIce::DArray          # TAUCLI - Optical depth, ice-path (-)
        tauLiq::DArray          # TAUCLW - Optical depth, liquid-water-path (-)
        potVort::DArray         # PV - Ertel potential vorticity (K.m2/kg/s)
        relHum::DArray          # RH - Relative humidity (%, read as unitless)
        specHum::DArray         # QV - Specific humidity (g/kg, read as kg/kg))
        liqRatio::DArray        # QL - Liquid water mixing ratio (kg/kg)
        iceRatio::DArray        # QI - Ice water mixing ratio (kg/kg)
        UWind::DArray           # U - Zonal (E-W) wind (m/s)
        VWind::DArray           # V - Meridional (N-S) wind (m/s)
        pVelocity::DArray       # OMEGA - Vertical pressure velocity (Pa/s)
        TMet::DArray            # T - Temperature (K)
        # Derived fields
        airMass::DArray         # AD - Mass of air in gridbox (kg/box)
        airDens::DArray         # AIRDEN - Air mass density (kg/m3)
        airVol::DArray          # AIRVOL - Air volume (m3/box)
        H2Ovv::DArray           # AVGW - H2O mixing ratio (v/v)
        dzBox::DArray           # BXHEIGHT - Vertical extent (m/box)
        dpBox::DArray           # DELP - Delta-P extent (hPa/box)
        cosSZA::DArray          # SUNCOS - Cosine(SZA) at current time 
        cosSZAMid::DArray       # SUNCOS_MID - Cosine(SZA) at midpoint of current chem timestep
        cosSZAOld::DArray       # SUNCOS_MID_5hr - Cosine(SZA) at midpoint of chem timestep 5h ago
        cldTop::DArray          # Derived - cloud tops (2D, CTM level)
        # Metadata
        I6File::IOStream
        nI6Fields::Int
        lastI6Date::Int
        lastI6Time::Int
        lastI6FileTime::TimeVecType
        lastI6ReadTime::TimeVecType
        A6File::IOStream
        nA6Fields::Int
        lastA6Date::Int
        lastA6Time::Int
        lastA6FileTime::TimeVecType
        lastA6ReadTime::TimeVecType
        A3File::IOStream
        nA3Fields::Int
        lastA3Date::Int
        lastA3Time::Int
        lastA3FileTime::TimeVecType
        lastA3ReadTime::TimeVecType
end

# Directory options
type JZDirType
    metDir::ASCIIString
    outputDir::ASCIIString
    machineFile::ASCIIString
end

# Options read/derived from JZInput
type JZOptType
    # Simulation name
    simName::ASCIIString
    # Directory for GEOS-5 data at 4x5
    metDir::ASCIIString
    # Directory for output
    outDir::ASCIIString
    # Machine file
    machineFile::ASCIIString
    # Start/end dates (YYYYMMDD)
    dStart::Int
    dEnd::Int
    # Start/end times (HHMMSS)
    tStart::Int
    tEnd::Int
    # Timestep length (mins)
    tStep::Int
    # Tracer data
    tracerNames::Array
    molWeight::Array
    # Debug mode
    isDebug::Bool
    maxOutput::Bool
end

# Timing data
type JZTimeType
    # All dates and times are GMT
    # Dates are integers formatted as YYYYMMDD
    # Times are integers formatted as HHMMSS
    # "Tau" times are floating point numbers, specifically the number of hours since 1985-01-01 00:00:00
    tStart::TimeVecType
    tCurr::TimeVecType
    tEnd::TimeVecType
    tauStart::FloatingPoint
    tauCurr::FloatingPoint
    tauEnd::FloatingPoint
    tStepDyn::Int # Minutes
    tElapsed::Int # Minutes
end

# Grid specification
type JZGridType
    # Straight definitions of longitude and latitude edges (deg, rad)
    lonEdge::Array
    latEdge::Array
    lonEdgeR::Array
    dLonR::FloatingPoint
    latEdgeR::Array
    dLatR::Array

    # Straight definitions of longitude and latitude centers (deg, rad)
    lonCenter::Array
    latCenter::Array
    lonCenterR::Array
    latCenterR::Array

    # Sine of the above
    sinLatEdge::Array
    dSinLatEdge::Array
    sinLonEdge::Array
    sinLatCenter::Array
    sinLonCenter::Array

    # d(sin(lat))/d(lat) ~ cos(latCenter)
    cosLatWeighted::Array
    cosLatEdge::Array

    # Temporal gradients wrt lat/lon
    dtdx5::Array
    dtdy5::Array

    # Simple grid definitions
    nLon::Int
    nLat::Int
    nLev::Int

    # Extent of the polar caps
    nCap::Int

    # Arrays to help creation of distributed arrays
    chunk2D::Array
    chunk3D::Array

    # Grid area in square meters and advection factors
    gridArea::DArray
    gridRel::DArray
    geoFactor::Array
    capFactor::FloatingPoint

    # Offset and multiplier data for pressure calculations
    pOffset::Array
    pFactor::Array
    deltaPOff::Array
    deltaPFac::Array
end

# Constants
type JZConstDataType
    rEarth::FloatingPoint   # Radius of earth (m)
    avogNum::FloatingPoint  # Avogadro's number (molec/mol)
    RUniv::FloatingPoint    # Universal gas constant (J/K/mol)
    airMW::FloatingPoint    # Molecular mass of air (kg/kmol == g/mol)
    gSurf::FloatingPoint    # Grav. acc. at surface of earth (m/s2)
end

type JZTracGridType
    mixRatio::DArray        # Mixing ratios (v/v)
    phaseRatio::DArray      # Phase ratios (gas - liquid - solid)
    molWeight::Array        # MW tracer (g/mol)
    relWeight::Array        # MW air / MW tracer
    nTracers::Int           # Number of tracers
    tracerNames::Array      # Tracer names
end
# ---END TYPE DEFINITIONS---

# ---START TYPE CONSTRUCTORS---
function JZTracGridType(optData::JZOptType,gridData::JZGridType,constData::JZConstDataType)
    # FUNCTION NAME     : JZTracGridType
    # DESCRIPTION       : Tracer grids and related arrays
    # INPUTS:
    #   optData         : J-Z options data
    #   gridData        : J-Z grid data
    #   constData       : J-Z constants
    # OUTPUTS:
    #   JZTracData      : Tracer grids
    #
    # ---START CODE---
    # Establish the output grids
    nLon = gridData.nLon
    nLat = gridData.nLat
    nLev = gridData.nLev
    nTracers = length(optData.tracerNames)
    tracerNames = optData.tracerNames
    molWeight = optData.molWeight
    relWeight = constData.airMW./molWeight
    # Each processor handles all tracers but only their own domain
    chunk3DTrac = [gridData.chunk3D...,1]
    mixRatio = dfill(0.0,(nLon,nLat,nLev,nTracers),workers(),chunk3DTrac)
    # Phase fraction (gas-liquid-solid) of each tracer
    # Unlikely to use this, but futureproofing in any case
    chunk3DPhase = [chunk3DTrac...,1]
    phaseRatio = dfill(0.0,(nLon,nLat,nLev,nTracers,3),workers(),chunk3DPhase)
    return JZTracGridType(  mixRatio,   phaseRatio, molWeight,  
                            relWeight,  nTracers,   tracerNames )
end

function JZPGridType(pSurf::DArray,gridData::JZGridType)
    # FUNCTION NAME     : JZPGridType
    # DESCRIPTION       : Calculates grid box edge and center pressures
    #                     relevant to the given surface pressure
    # INPUTS:
    #   pOffset         : Ap in GEOS-Chem
    #   pFactor         : Bp in GEOS-Chem
    #   pSurf           : 2-D distributed array
    #   gridData        : Resolution specification
    # OUTPUTS:
    #   JZPGridType     : Pressure grids
    #
    # ---START CODE---
    pEdge = dfill(0.0,(gridData.nLon,gridData.nLat,gridData.nLev+1),workers(),gridData.chunk3D)
    pCenter = dfill(0.0,(gridData.nLon,gridData.nLat,gridData.nLev),workers(),gridData.chunk3D)
    pOffset = gridData.pOffset
    pFactor = gridData.pFactor
    nLev = gridData.nLev
    @sync begin
        [@spawnat iWorker begin
            # Retrieve the pressure grid calculation data
            pOffset = remotecall_fetch(myid(),()->pOffset)
            pFactor = remotecall_fetch(myid(),()->pFactor)
            nLev = remotecall_fetch(myid(),()->nLev)
            pEdgeLocal = localpart(pEdge)
            pCenterLocal = localpart(pCenter)
            pSurfLocal = localpart(pSurf)
            localIdx = myindexes(pSurf)
            # Process each column independently
            for iLon = 1:length(localIdx[1])
                for iLat = 1:length(localIdx[2])
                    pSurfCell = pSurfLocal[iLon,iLat]
                    pEdgeLocal[iLon,iLat,1] = pSurfCell
                    for iLev = 2:nLev+1
                        pEdgeLocal[iLon,iLat,iLev] = (pSurfCell*pFactor[iLev]) + pOffset[iLev]
                        pCenterLocal[iLon,iLat,iLev-1] = sum(pEdgeLocal[iLon,iLat,(iLev-1):iLev])/2.0
                    end
                end
            end
        end
        for iWorker in workers()]
    end
    return JZPGridType(pEdge,pCenter)
    # ---END CODE---
end

function JZConstDataType()
    # FUNCTION NAME     : JZConstDataType
    # DESCRIPTION       : Holds constants for J-Zephyr. Establishing
    #                     this once and passing it around is ensures
    #                     consistency between functions
    # INPUTS:
    #   jDay            : Astronomical Julian day
    # OUTPUTS:
    #   jVec            : TimeVecType time vector
    #
    # ---START CODE---
    rEarth = 6.375e6            # Radius of earth (m)
    avogNum = 6.02214129e23     # Avogadros' number (molec/mol)
    RUniv = 8.3144622           # Universal gas constant (J/K/mol)
    airMW = 28.97               # Molecular weight of air (kg/kmol)
    gSurf = 9.8                 # Gravitational acceleration at the surface (m/s2)
    return JZConstDataType(rEarth,avogNum,RUniv,airMW,gSurf)
    # ---END CODE---
end

function JZGridType(optData::JZOptType,constData::JZConstDataType)
    # FUNCTION NAME     : JZGridType
    # DESCRIPTION       : Calculates grid variables relevant to the
    #                     chosen model resolution. Currently hardcoded
    #                     to generate a 4x5 grid
    # INPUTS:
    #   optData         : J-Zephyr options structure
    #   constData       : Constants relevant to J-Z
    # OUTPUTS:
    #   JZGridType      : Grid specification for J-Zephyr
    #
    # ---START CODE---
    # For now, we're just ignoring the data and setting up a default 4x5 grid
    lonEdge = float([-182.5:5:177.5])
    latEdge = float([-90,-88:4:88,90])
    lonCenter = lonEdge[1:end-1] + (0.5.*diff(lonEdge))
    latCenter = latEdge[1:end-1] + (0.5.*diff(latEdge))
    nLon = length(lonCenter)
    nLat = length(latCenter)
    nLev = 72
    nCap = 2

    # Extract the 2D configuration
    gridArea = dfill(0.0,(nLon,nLat),workers())
    chunk2D = [size(gridArea.chunks)...]
    chunk3D = [size(gridArea.chunks)...,1]

    # Set up advection factors at the same time
    gridRel = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    # Geometrical factor for meridional advection using
    # correct spherical geometery
    geoFactor = fill(0.0,nLat)
    capFactor = 0.0 # Special case (polar cap)
    gridMult = 2.0*pi*(constData.rEarth*constData.rEarth)/float(nLon)
    gridLocal = zeros(nLon,nLat)
    for iLat = 1:nLat
        ABox = gridMult * (sind(latEdge[iLat+1]) - sind(latEdge[iLat])) 
        gridLocal[:,iLat] = ABox
    end

    # Relative area
    sumArea = sum(gridLocal)
    relLocal = gridLocal./sumArea

    # Distribute the arrays
    @sync begin
        [@spawnat iWorker begin
            coreArea = localpart(gridArea)
            coreRel = localpart(gridRel)
            targIdx = myindexes(gridArea)
            coreArea[:] = remotecall_fetch(myid(),()->gridLocal[targIdx...])
            coreRel[:] = remotecall_fetch(myid(),()->relLocal[targIdx...])
        end
        for iWorker in workers()]
    end

    # Geometrical factors for meridional advection
    dPFac = pi./(nLat-1) # ?!
    geoFactor = dPFac./(2.0.*relLocal[1,:].*nLon)
    capFactor = dPFac./(2.0.*sum(relLocal[1,1:2]).*nLon)

    # The following quantities are required for the transport core
    lonEdgeR = lonEdge.*pi./180.0
    latEdgeR = latEdge.*pi./180.0
    dLonR = 2.0*pi./nLon
    dLatR = diff(latEdgeR)

    # Need timestep in seconds
    tStepSec = float(optData.tStep)*60.0

    # Now the centers
    dLon = 360.0/nLon
    dLat = diff(latEdge)
    lonCenter = lonEdge[1:end-1] + (0.5.*dLon)
    latCenter = latEdge[1:end-1] + (0.5.*dLat)
    lonCenterR = lonCenter.*pi./180
    latCenterR = latCenter.*pi./180

    sinLatEdge = sin(latEdgeR)
    sinLatEdge[1] = -1.0
    sinLatEdge[end] = 1.0

    # Special treatment at the poles
    # Polar cap spans two latitude bands
    dLatR[1] = 2.0*dLatR[1]
    dLatR[end] = 2.0*dLatR[end]

    sinLatCenter = sin(latCenterR)
    # These are wrong, but this is what's
    # specified in GEOS-Chem...
    sinLatCenter[1] = -1.0
    sinLatCenter[end] = 1.0

    cosLatEdge = cos(latEdgeR)
    cosLatEdge[1] = 0.0
    cosLatEdge[end] = 0.0
    
    sinLonEdge = sin(lonEdgeR)
    sinLonCenter = sin(lonCenterR)
    sinLonCenter[1] = -1.0

    # This term is "gw" in GEOS-Chem
    dSinLatEdge = diff(sinLatEdge)

    # d(sin(lat))/d(lat) == mid(cos(lat))
    # Note that this is NOT equal to cos(midlat)!
    # This term is "cosp" in GEOS-Chem
    cosLatWeighted = dSinLatEdge./dLatR

    dtdx5 = 0.5 .* tStepSec ./ (dLonR .* constData.rEarth .* cosLatWeighted)
    dtdy5 = 0.5 .* tStepSec ./ (dLatR .* constData.rEarth)

    # Pressure offset (AP)
    pOffset =   [0.000000e+00, 4.804826e-02, 6.593752e+00, 1.313480e+01,
                 1.961311e+01, 2.609201e+01, 3.257081e+01, 3.898201e+01,
                 4.533901e+01, 5.169611e+01, 5.805321e+01, 6.436264e+01,
                 7.062198e+01, 7.883422e+01, 8.909992e+01, 9.936521e+01,
                 1.091817e+02, 1.189586e+02, 1.286959e+02, 1.429100e+02,
                 1.562600e+02, 1.696090e+02, 1.816190e+02, 1.930970e+02,
                 2.032590e+02, 2.121500e+02, 2.187760e+02, 2.238980e+02,
                 2.243630e+02, 2.168650e+02, 2.011920e+02, 1.769300e+02,
                 1.503930e+02, 1.278370e+02, 1.086630e+02, 9.236572e+01,
                 7.851231e+01, 6.660341e+01, 5.638791e+01, 4.764391e+01,
                 4.017541e+01, 3.381001e+01, 2.836781e+01, 2.373041e+01,
                 1.979160e+01, 1.645710e+01, 1.364340e+01, 1.127690e+01,
                 9.292942e+00, 7.619842e+00, 6.216801e+00, 5.046801e+00,
                 4.076571e+00, 3.276431e+00, 2.620211e+00, 2.084970e+00,
                 1.650790e+00, 1.300510e+00, 1.019440e+00, 7.951341e-01,
                 6.167791e-01, 4.758061e-01, 3.650411e-01, 2.785261e-01,
                 2.113490e-01, 1.594950e-01, 1.197030e-01, 8.934502e-02,
                 6.600001e-02, 4.758501e-02, 3.270000e-02, 2.000000e-02,
                 1.000000e-02 ]

    # Pressure factor (BP)
    pFactor =   [1.000000e+00, 9.849520e-01, 9.634060e-01, 9.418650e-01,
                 9.203870e-01, 8.989080e-01, 8.774290e-01, 8.560180e-01,
                 8.346609e-01, 8.133039e-01, 7.919469e-01, 7.706375e-01,
                 7.493782e-01, 7.211660e-01, 6.858999e-01, 6.506349e-01,
                 6.158184e-01, 5.810415e-01, 5.463042e-01, 4.945902e-01,
                 4.437402e-01, 3.928911e-01, 3.433811e-01, 2.944031e-01,
                 2.467411e-01, 2.003501e-01, 1.562241e-01, 1.136021e-01,
                 6.372006e-02, 2.801004e-02, 6.960025e-03, 8.175413e-09,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00,
                 0.000000e+00 ]

    # Need deltas calculated a specific way for advection calculations
    deltaPOff = pOffset[1:end-1] - pOffset[2:end]
    deltaPFac = pFactor[1:end-1] - pFactor[2:end]

    return JZGridType(  lonEdge,        latEdge,        lonEdgeR,       dLonR,
                        latEdgeR,       dLatR,          lonCenter,      latCenter,
                        lonCenterR,     latCenterR,     sinLatEdge,     dSinLatEdge,
                        sinLonEdge,     sinLatCenter,   sinLonCenter,   cosLatWeighted,
                        cosLatEdge,     dtdx5,          dtdy5,          nLon,           
                        nLat,           nLev,           nCap,         
                        chunk2D,        chunk3D,        
                        gridArea,       gridRel,        geoFactor,      capFactor,
                        pOffset,        pFactor,        deltaPOff,      deltaPFac)
end

function TimeVecType(jDay::FloatingPoint)
    # FUNCTION NAME     : TimeVecType
    # DESCRIPTION       : Converts an astronomical Julian date to
    #                     a time vector
    # INPUTS:
    #   jDay            : Astronomical Julian day
    # OUTPUTS:
    #   jVec            : TimeVecType time vector
    #
    # ---START CODE---
    jDOff = jDay + 0.5
    jDayInt = itrunc(jDOff)
    jDayFrac = jDOff - itrunc(jDayInt)

    if jDayInt > 2299160.0
        ATemp = itrunc((jDayInt - 1867216.25)/36524.25)
        BTemp = jDayInt + 1 + ATemp - itrunc(ATemp/4)
    else
        BTemp = jDayInt
    end
    CTemp = BTemp + 1524.0
    DTemp = itrunc((CTemp-122.1)/365.25)
    ETemp = itrunc(365.25*DTemp)
    GTemp = itrunc((CTemp-ETemp)/30.6001)
    jDayFloat = CTemp - ETemp + jDayFrac - itrunc(30.6001*GTemp)
    jDayFrac = jDayFloat - itrunc(jDayFloat)
    if GTemp < 13.5
        jMonth = GTemp - 1.0
    else
        jMonth = GTemp - 13.0
    end
    if jMonth > 2.5
        jYear = DTemp - 4716.0
    else
        jYear = DTemp - 4715.0
    end
    jHourFloat = jDayFrac*24.0
    jMinFloat = (jHourFloat-itrunc(jHourFloat))*60.0
    jSecFloat = int((jMinFloat-itrunc(jMinFloat))*60.0)
    return TimeVecType(itrunc(jYear),itrunc(jMonth),itrunc(jDayFloat),itrunc(jHourFloat),itrunc(jMinFloat),int(jSecFloat))
    # ---END CODE---
end

function TimeVecType(year::Int,month::Int,day::Int)
    # Default in the case of no HHMMSS spec
    return TimeVecType(year,month,day,0,0,0)
end

function TimeVecType(nYMD::Int,nHMS::Int=0)
    # Autoconversion constructor
    (year,month,day) = parseIntDT(nYMD)
    (hour,minute,second) = parseIntDT(nHMS)
    return TimeVecType(year,month,day,hour,minute,second)
end

function JZTimeType(optData::JZOptType)
    # Calculate each of the three times
    tStart = TimeVecType(optData.dStart,optData.tStart)
    tEnd = TimeVecType(optData.dEnd,optData.tEnd)
    tauStart = calcTau(tStart)
    tauEnd = calcTau(tEnd)
    return JZTimeType(  tStart,             tStart,         tEnd,
                        tauStart,           tauStart,       tauEnd,
                        optData.tStep,      0                               )
end

# JZMetType constructors
function JZMetType(gridData::JZGridType)
    return JZMetType(gridData.nLon,gridData.nLat,gridData.nLev,gridData.chunk2D,gridData.chunk3D)
end

function JZMetType(nLon::Int,nLat::Int,nLev::Int,chunk2D::Array,chunk3D::Array)
    #LWI = dfill(0.0,(nLon,nLat),workers())
    ## Extract the 2D configuration
    #chunk2D = [size(LWI.chunks)...]
    #chunk3D = [size(LWI.chunks)...,1]
    # Now the rest
    LWI = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    pSurf = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    pSurfStart = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    pSurfEnd = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    pSeaLevel = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    O3DUCol = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    O3DUColStart = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    O3DUColEnd = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    O3DUTrop = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    # A-3 fields
    sfcAlbedo = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    cldFrc2D = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    evapFlux = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    greenIdx = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    wetRoot = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    wetTop = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    sensFlux = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    LAI = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    PARDiffuse = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    PARDirect = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    PBLDepth = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    prcpAnvSurf = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    prcpTotSurf = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    prcpConSurf = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    prcpSnoSurf = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    netRadLW = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    upRadLW = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    specHum2m = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    netRadSW = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    snowGeo = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    snowH2O = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    pTrop = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    T2m = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    TSkin = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    U10m = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    UStar = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    V10m = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    Z0 = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    EFlux = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    fracLand = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    fracOcean = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    fracLake = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    fracLandIce = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    # A-6 fields
    cldFrc3D = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    detrainFlux = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    # Convective flux is edge-defined
    convectFlux = dfill(0.0,(nLon,nLat,nLev+1),workers(),chunk3D)
    iceTendMoist = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    liqTendMoist = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    vapTendMoist = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    prcpConForm = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    prcpLrgForm = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    vapTend = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    tauCld = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    tauIce = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    tauLiq = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    potVort = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    relHum = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    specHum = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    liqRatio = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    iceRatio = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    UWind = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    VWind = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    pVelocity = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    TMet = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    airMass = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    airDens = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    airVol = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    H2Ovv = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    dzBox = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    dpBox = dfill(0.0,(nLon,nLat,nLev),workers(),chunk3D)
    cosSZA = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    cosSZAMid = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    cosSZAOld = dfill(0.0,(nLon,nLat),workers(),chunk2D)
    cldTop = dfill(0,(nLon,nLat),workers(),chunk2D)
    # Metadata
    I6File=IOStream("I6")
    nI6Fields=0
    lastI6Date=-1
    lastI6Time=-1
    lastI6FileTime = TimeVecType(0,0,0,0,0,0)
    lastI6ReadTime = TimeVecType(0,0,0,0,0,0)
    A6File=IOStream("A6")
    nA6Fields=0
    lastA6Date=-1
    lastA6Time=-1
    lastA6FileTime = TimeVecType(0,0,0,0,0,0)
    lastA6ReadTime = TimeVecType(0,0,0,0,0,0)
    A3File=IOStream("A3")
    nA3Fields=0
    lastA3Date=-1
    lastA3Time=-1
    lastA3FileTime = TimeVecType(0,0,0,0,0,0)
    lastA3ReadTime = TimeVecType(0,0,0,0,0,0)
    return JZMetType(   LWI,            pSurf,          pSurfStart,     pSurfEnd,
                        pSeaLevel,      O3DUCol,        O3DUColStart,   O3DUColEnd,     O3DUTrop,
                        sfcAlbedo,      cldFrc2D,       evapFlux,       greenIdx,       wetRoot,
                        wetTop,         sensFlux,       LAI,            PARDiffuse,     PARDirect,
                        PBLDepth,       prcpAnvSurf,    prcpTotSurf,    prcpConSurf,    prcpSnoSurf,
                        netRadLW,       upRadLW,        specHum2m,      netRadSW,       snowGeo,
                        snowH2O,        pTrop,          T2m,            TSkin,          U10m,
                        UStar,          V10m,           Z0,             EFlux,          fracLand,
                        fracOcean,      fracLake,       fracLandIce,    cldFrc3D,       detrainFlux,
                        convectFlux,    iceTendMoist,   liqTendMoist,   vapTendMoist,   prcpConForm,
                        prcpLrgForm,    vapTend,        tauCld,         tauIce,         tauLiq,
                        potVort,        relHum,         specHum,        liqRatio,       iceRatio,
                        UWind,          VWind,          pVelocity,      TMet,           
                        airMass,        airDens,        airVol,         H2Ovv,          dzBox,
                        dpBox,          cosSZA,         cosSZAMid,      cosSZAOld,      cldTop,
                        I6File,         nI6Fields,      lastI6Date,     lastI6Time,     lastI6FileTime,
                        lastI6ReadTime, A6File,
                        nA6Fields,      lastA6Date,     lastA6Time,     lastA6FileTime, lastA6ReadTime,
                        A3File,         nA3Fields,
                        lastA3Date,     lastA3Time,     lastA3FileTime, lastA3ReadTime                 )
end

# ---END TYPE CONSTRUCTORS---
